.TH WHICH 1 \" Apr\ 04\ 1990
.SH NAME
which \- give alias, function or path expansion of command
.SH SYNOPSIS
.B which
[
.B \-i
] [
.B \-a
] [
.B \-\-
] [
.I command
]
.SH DESCRIPTION
.B Which
provides the user with the full expansion of the
.I command
argument, be it either an
.IR alias ,
a
.IR "shell function" ,
or an executable file (default).
To enable search for
.I aliases
and
.I "shell functions"
the user should supply the
.B \-i
(interactive) flag.
In that case
.B which
expects as standard input the expansion of the
.I alias
or
.IR "shell function" .
If the standard input is empty or the
.B \-i
flag has not been given,
.B which
will try to locate
.I command
in the user's
.BR PATH .
The interactive mode is easily used by setting an
.I alias
like the following:
.sp
.RS
.ft B
.nf
alias which alias !\\$ \\| /usr/local/bin/which \-i !\\*
.fi
.RE
.ft R
.sp
in
.IR csh ,
or
.sp
.ft B
.RS
.nf
alias which eval alias '\\"\\$$#\\" |' \\
	/usr/local/bin/which \-i '${1+"$@"}'
.fi
.RE
.ft R
.sp
in shells which are supersets of
.I sh
and which know
.IR aliases .
If your shell has
.IR "shell functions" ,
you can use the following function:
.sp
.ft B
.RS
.nf
which()
{
	eval last=\\"\\$$#\\"
	set | sed \-n "/^$last(){$/,/^}$/p" |
		/usr/local/bin/which \-i ${1+"$@"}
}
.fi
.RE
.ft R
.sp
If the
.B \-a
(all) flag is given,
.B which
will not stop after the first `match', but search for all occurrences of
.I command
in the user's
.BR PATH .
The
.B \-\-\
flag can be used to end the list of options: the next argument (if present)
will be taken as
.IR command ,
even if it starts with a `\-'.
.B "which [\-i] [\-a] [\-\-]"
without further arguments prints the user's
.B PATH
broken up into its components,
one per line.
.PP
This new version of the
.B which
command is not a
.I csh
script.
Being an executable it is much faster, and not sourcing
.I .cshrc
it gives a true picture of one's
.I aliases.
Furthermore it will give the correct answers even if:
.IP \-
the
.I "effective uid"
.RI ( gid )
differs from the
.I real
uid (gid)
.IP \-
the effective uid is 0
.RI ( root ).
.SH EXAMPLE
.PD
.ft B
.nf
% alias
which	alias !$ | /usr/local/bin/which \-i !*
% which which
which	alias !$ | /usr/local/bin/which \-i !*
% which \-a which
which	alias !$ | /usr/local/bin/which \-i !*
/usr/local/bin/which
/usr/ucb/which
%
.fi
.ft R
.SH AUTHOR
Maarten Litmaath @ VU Informatika Amsterdam
